#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
file: pulse_height.py 
discription: 
"""
import os
import argparse
import datetime
import time
import logging
import dask.dataframe as dd
from dask.distributed import LocalCluster
import pandas as pd
import numpy as np
import toml

from .pulse_height import pulse_height
from .cross_correlation import cross_correlation

def main():
    parser = argparse.ArgumentParser(
                    prog='pyMPPost',
                    description='What the program does',
                    epilog='Text at the bottom of help')
    
    parser.add_argument('input_file')
    args = parser.parse_args()
    
    with open(args.input_file) as toml_file:
        input_toml = toml.load(toml_file) #Loads toml file into a python dict
    input_toml["input_file"] = args.input_file

    logging.basicConfig(filename=f"{input_toml['i/o']['output_root']}.log", filemode='w+', 
                        format='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S %p', 
                        level=logging.DEBUG)
    
    
    dask_cluster = LocalCluster(silence_logs=logging.ERROR)
    
    _mppost_start(input_toml, dask_cluster.get_client())
    

def _mppost_start(info, client):
    # Save Start time
    run_start_time = time.time()
    logging.info("Beginning pyMPPost run...")
      
    detect_map, cell_to_mats = _detector_data(info)
    ### Read in Polimi Data to pyMPPost Object ###
    
    plm_ddf = _read_polimi(info, detect_map, cell_to_mats, client) 
    
    ### Run MPPost Modules ###
    results = {}
    
    # Check if Pulse Height Module On
    if info["pulse height"]["module_on"]:
        PH_df, results["pulse height"] = pulse_height(plm_ddf, info, client)
        
    if info["cross correlation"]["module_on"]:
        cross_correlation(PH_df, info)
    
    _gen_summary(plm_ddf, info, client, results, run_start_time)
    
def _detector_data(info):
    cell_to_mats = {}
    detect_map = {}
    
    for (material, cells) in zip(info["detectors"]["detector_mats"],info["detectors"]["detector_cells"]):
            cell_to_mats[cells[0]] = material
            if len(cells) > 1:
                for cell in cells[1:]:
                    detect_map[cell] = cells[0]
                
    mats = []
    for mat_path in info["detectors"]["material_list"]:
        with open(mat_path) as mat_toml_file:
            mat = toml.load(mat_toml_file)
        mats.append(mat)
    info["mats"] = mats
    
    for mat in mats:
        if mat["mat_type"] == 1:
            mat_birks = mat["calibration"].pop("birks")
            with open(mat_birks["stopping_power"]) as dEdx_file:
                lines = dEdx_file.readlines()
                length = len(lines)
                energy = np.zeros((length,))
                dEdx = np.zeros((length,))
                intergrand = np.zeros((length,))
                for i, line in enumerate(lines):
                    line = line.split()
                    energy[i] = float(line[0])
                    # For stilbene files, only the first and the 3rd line are read in
                    if (os.path.basename(mat_birks["stopping_power"]) == "stilbene_dEdx.txt"):
                        dEdx[i] = float(line[2])
                    else:
                        dEdx[i] = float(line[1])
            energy = np.insert(energy, 0, values=0.0)
            dEdx = np.insert(dEdx, 0, values=0.0)
            intergrand = mat_birks["S"]/(1 + mat_birks["kB"]*dEdx)
            length = len(intergrand)
            birks_light = np.zeros((length,))
            for i in range (length):
                birks_light[i] = np.trapz(y=intergrand[0:i+1], x=energy[0:i+1]) 
            mat["calibration"]["birks_energy"] = energy
            mat["calibration"]["birks_light"] = birks_light
          
    detectors = {}
    for cell in cell_to_mats:
       detectors[cell] = mats[cell_to_mats[cell]]
    return detect_map, cell_to_mats
        
def _read_polimi(info, detect_map, cells_to_mats, client):
    """
    Stores Polimi Data into a Dask Dataframe (plm_ddf) with index on history.
    
    Dataframe Info
        Index: history (int) 
        
        Columns:
            particle_num (int), particle_type (int), interaction_type (int), target_nucleus (int), cell (int), energy_deposited (float), time (float), 
            x-pos (float), y-pos (float), z-pos (float), weight (float), generation_num (int), num_scatters (int), code (int), prior_energy (float) 
    """ 
    
    logging.info(f"Reading in MCNP-Polimi file: {info['i/o']['polimi_det_in']}")
    polimi_collision_file_dtypes = {
        "history":int, 
        "particle_num":int, 
        "particle_type":int,
        "interaction_type":int, 
        "target_nucleus":int, 
        "cell":int, 
        "energy_deposited":float, 
        "time":float, 
        "x-pos":float, 
        "y-pos":float, 
        "z-pos":float, 
        "weight":float, 
        "generation_num":int,
        "num_scatters":int, 
        "code":int, 
        "prior_energy":float
    }
    plm_ddf = dd.read_csv(f"{info['i/o']['polimi_det_in']}", header=None, skipinitialspace=True, sep="\s+",  
                            names=["history", "particle_num", "particle_type","interaction_type", "target_nucleus", "cell", 
                                    "energy_deposited", "time", "x-pos", "y-pos", "z-pos", "weight", "generation_num",
                                    "num_scatters", "code", "prior_energy"], dtype=polimi_collision_file_dtypes).set_index("history", sorted=True)
    if detect_map:
        plm_ddf["cell"] = plm_ddf["cell"].replace(detect_map)
    
    plm_ddf["mat_num"] = plm_ddf["cell"].replace(cells_to_mats)
    
    meta = plm_ddf._meta
    plm_ddf = plm_ddf.map_partitions(lambda df: df.sort_values(by=["history", "cell", "time"]), meta=meta)
    
    return client.persist(plm_ddf)

def _gen_summary(plm_ddf, info, client, results, start_time):
    # Open Summary File
    sum_of = open(f"{info['i/o']['output_root']}.txt", mode="w")
    sum_of.write("Post Processor Output File Summary\n")
    
    date_and_time = datetime.datetime.now()
    sum_of.write(f"Title: {info['title']}\n")      
    sum_of.write(f"Input File: {info['input_file']}\n")  
    sum_of.write(f"User: {info['username']}\n")      
    sum_of.write(f"Processed: {date_and_time}\n\n")  
    
    sum_of.write("MCNP-PoliMi File Characteristics\n")
    sum_of.write("--------------------------------\n")
    sum_of.write(f"Number of lines: {len(plm_ddf)}\n")
    sum_of.write(f"Number of histories: {client.compute(plm_ddf.index.nunique()).result()}\n\n")   
    
    # Check if Pulse Height Module On
    if info["pulse height"]["module_on"]:
        # Write Pulse Height Info to Summary File
        (total_num_pulses, neutron_num_pulses) = results["pulse height"]
        sum_of.write(f"Pulse Height Analysis\n")
        sum_of.write(f"---------------------\n")
        sum_of.write(f"Total number of pulses above threshold: {total_num_pulses}\n")
        sum_of.write(f"Total number of neutron pulses above threshold: {neutron_num_pulses}\n")   
        sum_of.write(f"Total number of photon pulses above threshold: {total_num_pulses - neutron_num_pulses}\n\n")
        
    sum_of.write(f"Runtime\n")    
    sum_of.write(f"-------\n")  
    sum_of.write(f"This run took {time.time() - start_time} seconds")
    sum_of.close()
    
    
if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
"""
file: cross_correlation.py 
discription: 
"""
import logging
import numpy as np
import pandas as pd

def cross_correlation(PH_df, info):
    logging.info("Beginning Cross Correlation")
    PH_df["particle_label"] = PH_df.particle_type.map({1:"n",2:"g", 3:"g"})
    start_detectors = info["cross correlation"]["start_detectors"]
    
    for start_detect in start_detectors:
        logging.info(f"Computing Cross Correlation for start detector {start_detect}")
        start_detect_mask = PH_df["cell"].eq(start_detect).groupby("history").transform("any")
        oth_detect_mask = PH_df["cell"].ne(start_detect).groupby("history").transform("any")
        filtered_df = PH_df[start_detect_mask & oth_detect_mask]
        cell_mask = (filtered_df["cell"] == start_detect)
        start_df = filtered_df[cell_mask]
        oth_df = filtered_df[~cell_mask]
        computed1 = np.split(start_df[["time","particle_label"]].values, np.unique(start_df.index.values, return_index=True)[1][1:])
        computed2 = np.split(oth_df[["time","particle_label"]].values, np.unique(oth_df.index.values, return_index=True)[1][1:])
        
        time_data = [np.subtract.outer(oth[:,0],start[:,0]).flatten() for start, oth in zip(computed1,computed2)]
        label_data = [np.add.outer(start[:,1], (oth[:,1])).flatten() for start, oth in zip(computed1,computed2)]
        time_diff = np.concatenate(time_data)
        labels = np.concatenate(label_data)
        
        CC_df = pd.DataFrame({"time_diff":time_diff, "labels":labels})
        if info["cross correlation"]["all_cc"] == "tsv":
            CC_df.to_csv(f"{info['i/o']['output_root']}_detector_{start_detect}_All_CC", index=False, sep="\t")
            
        total, edges = np.histogram(a=CC_df["time_diff"], bins=info["cross correlation"]["hist_bins"])
        nn, __ = np.histogram(a=CC_df[CC_df.labels == "nn"]["time_diff"], bins=edges)
        gn, __ = np.histogram(a=CC_df[CC_df.labels == "gn"]["time_diff"], bins=edges)
        ng, __ = np.histogram(a=CC_df[CC_df.labels == "nn"]["time_diff"], bins=edges)
        gg, __ = np.histogram(a=CC_df[CC_df.labels == "gg"]["time_diff"], bins=edges)
        np.savetxt(fname=f"{info['i/o']['output_root']}_detector_{start_detect}_CC_Hist", X=np.column_stack((edges[:-1],total, nn, gn, ng, gg)))
        
    logging.info("Cross Correlation Complete")


# -*- coding: utf-8 -*-
"""
file: pulse_height.py 
discription: 
"""
import logging
import numpy as np
from .cython_functs import calc_pulses


def pulse_height(plm_ddf, info, client):
    # Calculate Light
    meta = plm_ddf._meta.assign(light=[])
    plm_ddf = plm_ddf.map_partitions(lambda df: df.assign(light=_light_helper(df,info["mats"])),meta=meta)

    # Generate Pulses
    pgt_arr = []
    dt_arr = []
    
    for mat in info["mats"]:
        pgt_arr.append(mat["pulse_creation"]["pgt"]/ 10.0) 
        dt_arr.append(mat["pulse_creation"]["dt"]/ 10.0) 
        
    pgt_arr = np.array(pgt_arr)
    dt_arr = np.array(dt_arr)
    tol = info["pulse height"]["pulse_tol"]
    
    meta = meta.assign(pulse=[])
    pulses_q = plm_ddf.map_partitions(lambda df: df.assign(pulse=_pulse_helper(df, pgt_arr, dt_arr, tol)), meta=meta)
    
    # Sum Light by Pulse
    pulses_q = pulses_q.groupby(by=["history","cell","pulse"]).agg({'particle_type':'first', 'mat_num':'first', 'time': 'first', 'light': 'sum'})
    
    # Check and Run resolution sub-modules
    pulse_meta = pulses_q._meta
    if info["seed"] == 0:
        gen = np.random.default_rng()
    else:
        gen = np.random.default_rng(info["seed"])
            
    if info["pulse height"]["time_res_on"]:      
        # time resolution
        pulses_q = pulses_q.map_partitions(lambda df: df.assign(time=_time_res(df,info["mats"],gen)), meta=pulse_meta)
    
    if info["pulse height"]["energy_res_on"]:
        # energy resolution
        pulses_q = pulses_q.map_partitions(lambda df: df.assign(light=_energy_res(df,info["mats"], gen)), meta=pulse_meta)
    
    logging.info("Beginning Pulse Height Computation") 
    PH_df = client.compute(pulses_q).result().reset_index(["cell", "pulse"])
    PH_df = PH_df[PH_df["pulse"] >= 0]
    PH_df = PH_df[(PH_df["light"] >= info["pulse height"]["threshold"]["lower"]) & (PH_df["light"] <= info["pulse height"]["threshold"]["upper"])]
    logging.info("Pulse Height Computation Complete")  
    
    if info["pulse height"]["all_pulses"] == "tsv":
        PH_df.to_csv(f"{info['i/o']['output_root']}_All_Pulses", header=info['i/o']['output_headers'], sep="\t")
    elif info["pulse height"]["all_pulses"] == "parquet":
        PH_df.to_parquet(f"{info['i/o']['output_root']}_All_Pulses")
    
    logging.info("Pulse Height Export Complete") 
    
    total, edges = np.histogram(a=PH_df["light"], bins=info["pulse height"]["hist_bins"])
    neutrons, __ = np.histogram(a=PH_df[PH_df.particle_type == 1]["light"], bins=edges)
    photons, __ = np.histogram(a=PH_df[PH_df.particle_type == 2]["light"], bins=edges)
    np.savetxt(fname=f"{info['i/o']['output_root']}_Pulse_Hist", X=np.column_stack((edges[:-1],total,neutrons,photons)))
    
    return PH_df, (len(PH_df), len(PH_df[PH_df["particle_type"] == 1]))

def _light_helper(df, mats):
    conditions = []
    choices = []
    
    for i, mat in enumerate(mats):
        cal = mat["calibration"]
        mat_type = mat["mat_type"]
        
        conditions.extend([
            (df["mat_num"] == i) & (df["particle_type"] != 1),
            (df["mat_num"] == i) & (df["particle_type"] == 1) & ((df["target_nucleus"] == 6000) | (df["target_nucleus"] == 6012) | (df["target_nucleus"] == 6013)),
            (df["mat_num"] == i) & (mat_type == 1) & (df["particle_type"] == 1) & (df["target_nucleus"] == 1001),
            (df["mat_num"] == i) & (mat_type == 2) & (df["particle_type"] == 1) & (df["target_nucleus"] == 1001)
        ])
        choices.extend([
            df["energy_deposited"]*cal["photon"][0]+cal["photon"][1],
            df["energy_deposited"]*cal["carbon"],
            np.interp(df["energy_deposited"],cal["birks_energy"],cal["birks_light"]),
            df["energy_deposited"]*cal["photon"][0]+cal["photon"][1]
        ])
        
    return np.select(conditions,choices)

def _pulse_helper(df, pgt_arr, dt_arr, tol):
    return calc_pulses(df.index.values,df.cell.values, df.time.values, df.mat_num.values, pgt_arr, dt_arr, tol)

def _time_res(df,mats,gen):
    conditions = []
    choices = []
    for i, mat in enumerate(mats):
        conditions.append((df["mat_num"] == i))
        choices.append(mat["resolution"]["time"] / 10.0)
    time_res = np.select(conditions, choices)
    return gen.normal(loc=df["time"],scale=time_res)

def _energy_res(df, mats, gen):
    conditions = []
    choices = []
    for i, mat in enumerate(mats):
        mat_type = mat["mat_type"]
        en_res_p = mat["resolution"]["energy"]["photon"]
        if mat_type == 1:
            en_res_n = mat["resolution"]["energy"]["neutron"]
            conditions.append((df["mat_num"] == i) & (df["particle_type"] != 1))
            choices.append(df["light"] * np.sqrt(en_res_p[0]**2 + (en_res_p[1]**2/df["light"]) + (en_res_p[1]**2/df["light"]**2)))
            conditions.append((df["mat_num"] == i) & (df["particle_type"] == 1))
            choices.append(df["light"] * np.sqrt(en_res_n[0]**2 + (en_res_n[1]**2/df["light"]) + (en_res_n[1]**2/df["light"]**2)))
        else:
            conditions.append((df["mat_num"] == i))
            choices.append(df["light"] * np.sqrt(en_res_p[0]**2 + (en_res_p[1]**2/df["light"]) + (en_res_p[1]**2/df["light"]**2)))
            
    en_half_max = np.select(conditions, choices)
    en_sd = en_half_max / (2 * np.sqrt(2.0 * np.log(2)))        
    return gen.normal(loc=df["light"], scale= en_sd)


from setuptools import setup, Extension
import numpy

try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True

cmdclass = {}
ext_modules = []

if use_cython:
    ext_modules += [
        Extension("pyMPPost.cython_functs", ["src/pyMPPost/cython_functs.pyx"],define_macros=[('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')],
        include_dirs=[numpy.get_include()]),
    ]
    cmdclass.update({'build_ext': build_ext})
else:
    ext_modules += [
        Extension("pyMPPost.cython_functs", ["src/pyMPPost/cython_functs.c"]),
    ]

setup(
    name='pyMPPost',
    cmdclass=cmdclass,
    ext_modules=ext_modules,
)